// Jason Rohrer
// DownloadThread.java

/**
*
*	Thread for file download w/ status display GUI hooks
*
*	Created 5-11-2000
*	Mods:
*       Jason Rohrer    6-30-2000
*/

package konspire.client;

import konspire.common.Message;

import konspire.common.FileDescriptor;
import konspire.common.TwoWayMessageSocket;

import java.io.File;
import java.io.*;

/**
 * TransferThread for downloads.
 *
 * @author Jason Rohrer
 */
public class DownloadThread extends TransferThread {

    File filesDir;
	
	
	/**
	 * Construct an DownloadThread.
	 *
	 * @param fDes 			descriptor of file to download
	 * @param inFilesDir 	directory where files are located
	 *
	 */	
	public DownloadThread( FileDescriptor fDes, File inFilesDir ) {
		super( fDes );
		filesDir = inFilesDir;
		}
	
	boolean aborted = false;
	
	
	// the following functions implement the TransferThread abstract functions
	
	
	protected void startTransfer() {
		
		
		try {
			
			// decide if partial file exists locally
			
			long fileSize = f.getSize();
			
			long bytesRemain = fileSize;
			
			// create path to where file will be put when downloaded
			String filePath = filesDir.getPath() + File.separator + f.getName();
			
			// for file log
			String logFilePath = filePath + ".partial";
			
			// setup streams
			RandomAccessFile fileOutput;
			
			long bytesToSkip = 0;	// how many bytes are already in partial file (if any)?
			
			// look for log file (means partial download exists)
			File logFile = new File( logFilePath );
			if( logFile.exists() ) {	// log exists
				// resume download instead of starting new
				
				// stream for log file
				DataInputStream logFileInput = new DataInputStream (
						new FileInputStream( logFilePath ) );
				
				bytesToSkip = logFileInput.readLong();
				
				// make sure partial file is valid (matches file described by log
				File outFile = new File( filePath );
				if( outFile.exists() && bytesToSkip == outFile.length() ) {
					// partial file exists and matches log description
					fileOutput = new RandomAccessFile( filePath, "rw" );						
					// skip to end of partial file
					fileOutput.skipBytes( (int)bytesToSkip );
					bytesRemain -= bytesToSkip;
					}
				else {
					if( outFile.exists() ) {
						outFile.delete();
						}
					fileOutput = new RandomAccessFile( filePath, "rw" );
					}
				logFileInput.close();
				}
			else {	// no log file exists
				// delete any existing file
				File outFile = new File( filePath );
				if( outFile.exists() ) {
					outFile.delete();
					}
				fileOutput = new RandomAccessFile( filePath, "rw" );
				}
			
			
			// done initing partial output file
						

			TwoWayMessageSocket messageSock = new TwoWayMessageSocket( f.getHost() );
			
			// include # bytes to skip with file request
			messageSock.send( new ClientFileRequest( f, bytesToSkip ) );
		
			Message m = messageSock.receive();
			
			if( m instanceof FileTransferInit ) {
				FileTransferInit init = (FileTransferInit) m;
				if( ! init.getAccepted() ) {
					GuiMessage.show( "file download rejected by host: " + init.getStatus() );
					messageSock.close();
					fileOutput.close();
					return;
					}
				else {
					// transfer accepted by host
					
					int blockSize = init.getBlockSize();
					byte buff[] = new byte[ blockSize ];
					
					// start receiving file
					
					DataInputStream input = messageSock.getInputStream();
					
					
					// stream for log file
					DataOutputStream logFileOutput = new DataOutputStream (
								new FileOutputStream( logFilePath ) );
					
					try {
						while( bytesRemain > blockSize ) { // full blocks remain
							if( aborted ) {
								messageSock.close();
								throw new IOException( "Download aborted by user" );
								}
							input.readFully( buff );
							fileOutput.write( buff );
							bytesRemain -= blockSize;
							setProgress( fileSize - bytesRemain );
							}
						
						// last block may be a partial block
						if( bytesRemain > 0 ) {
							if( aborted ) {
								messageSock.close();
								throw new IOException( "Download aborted by user" );
								}
							input.readFully( buff, 0, (int)bytesRemain );
							fileOutput.write( buff, 0, (int)bytesRemain );
							bytesRemain = 0;
							setProgress( fileSize );
							}
						// done downloading file
						}
					catch( IOException ioe ) {
						GuiMessage.show( "partial file download: " + f.getName() );
						System.out.println( "Partial file download due to exception: " + ioe.toString() );
						
						transferFailed();
						
						// make sure socket is closed to free up connection on sending client
						messageSock.close();
						
						// file may have been downloaded partially
						fileOutput.close();
						
						File thisFile = new File( filePath );
						long bytesSoFar = thisFile.length();
						
						logFileOutput.writeLong( bytesSoFar );
						
						logFileOutput.close();
						return;
						}

					fileOutput.close();
					
					// file dl-ed successfully, no need for log file
					logFileOutput.close();
					logFile.delete();
					
					messageSock.close();
					return;	
					}
				
				}
			else {
				System.out.println( "Unexpected message type received from file host client: " + m.getClass().toString() );
				return;
				}
			
			}
		catch( IOException e ) {
			System.out.println( "Connection broken in download thread: " + e.toString() );
			transferFailed();
			}
		
		return;
		}



	protected void abortTransfer() {
		aborted = true;
		}

	}