// Jason Rohrer
// SearchPanel.java

/**
*
*	Bordered panel for search gui
*	Contains thread that retrieves search results
*
*	Created 5-11-2000
*	Mods:
*		Jason Rohrer	7-23-2000	Changed so that SearchThreads are set to 
*									minimum priority to make sure that the GUI
*									is always completely responsive.  Note that this
*									priority setting doesn't usually increase GUI responsiveness
*									noticibly, since TransferThreads are handling network connections
*									and are therefore blocked often enough so that the GUI thread
*									has ample time to run.  However, on faster network connections,
*									e.g., gigabit ethernet, the network threads will not
*									be blocked as often and could cause GUI responsiveness
*									to be sporadic.  We're assuming that as long as the GUI
*									isn't being manipulated or updated, MIN_PRIORITY thread
*									performance will not be affected.  This should be verified.
*/


package konspire.client;

import konspire.common.TwoWayMessageSocket;
import konspire.common.SearchQuery;
import konspire.common.SearchResults;
import konspire.common.Message;

import javax.swing.*;

import java.awt.FlowLayout;

import java.io.*;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

/**
 * Panel containing GUI controls for searching.
 *
 * @author Jason Rohrer
 */
public class SearchPanel extends EtchedTitledJPanel {


	String title = "search";
	
	TwoWayMessageSocket messageSock = null;		// where thread will do searches
	SearchResultsTable resultsTable;	// where results will be displayed
	
	Client parent;
	
	int searchFieldWidth = 20;
	
	JTextField searchField;
	
	JButton searchButton;
	
	JLabel resultsLabel;
	
	Object resultChoices[] = { new Integer(10), new Integer(20), new Integer(40) };
	
	JComboBox numResultsBox;
	
	JProgressBar progBar;
	
	
	JPanel topPanel;
	JPanel bottomPanel;
	
	ButtonActionListener buttonListen;
	
	/**
	 * Constructs a <code>SearchPanel</code>
	 *
	 * @param resTable table where search results will be displayed
	 * @par parent client of this search gui
	 */
	public SearchPanel( SearchResultsTable resTable, Client par ) {
		setTitle( title );
		
		resultsTable = resTable;
		
		parent = par;
		
		setLayout(new BoxLayout( this, BoxLayout.Y_AXIS ) );
		
		searchField = new JTextField( searchFieldWidth );
		
		searchButton = new JButton( "search" );
		
		resultsLabel = new JLabel( "results:" );
		
		numResultsBox = new JComboBox( resultChoices );
		
		progBar = new JProgressBar( 0, 3 );
		progBar.setStringPainted( true );
		progBar.setString( "not connected" );
		searchButton.setEnabled( false );
		
		
		resetProgress();
		
		
		topPanel = new JPanel( new FlowLayout( FlowLayout.CENTER ) );
		bottomPanel =  new JPanel( new FlowLayout( FlowLayout.CENTER ) );
		
		topPanel.add( searchField );
		topPanel.add( searchButton );
		
		
		bottomPanel.add( progBar );
		bottomPanel.add( resultsLabel );
		bottomPanel.add( numResultsBox );
		
		add( topPanel );
		add( bottomPanel );
		
		buttonListen = new ButtonActionListener();
		
		searchField.addActionListener( buttonListen );
		searchButton.addActionListener( buttonListen );
		
		}
	
	/**
	 * Sets socket that will be used for sending search requests
	 * and receiving search results.
	 *
	 * @param sock socket that will be used for search transmissions/receptions
	 */
	public void setSearchSocket( TwoWayMessageSocket sock ) {
		messageSock = sock;
		resetProgress();
		}
	
	/**
	 * Sets search progress bar back to starting value/message.
	 *
	 */
	private void resetProgress() {
		if( messageSock == null ) {
			progBar.setString( "not connected" );
			searchButton.setEnabled( false );
			}
		else {
			progBar.setString( "ready" );
			searchButton.setEnabled( true );
			}
		progBar.setValue( 0 );	
		}
	
	/**
	 * Starts search for a particular search string.
	 *
	 * @param searchString string to search for
	 * @param numResults number of results to request
	 */
	private void startSearch( String searchString, int numResults) {
		if( messageSock != null ) {
			SearchQuery sq = new SearchQuery( searchString, numResults );
			SearchThread sThread = new SearchThread( sq );
			searchButton.setEnabled( false );
			sThread.start();
			}
		}
	
	/**
	 * Callback from thread if connection breaks.
	 */
	private void connectionBroken() {
		parent.serverConnectionBroken();
		}
	
	/**
	 * Listener for buttons in the search GUI.
	 */
	private class ButtonActionListener implements ActionListener {
		public void actionPerformed( ActionEvent e ) {
			if( e.getSource() == searchButton || e.getSource() == searchField ) {
				Integer resultsInteger = (Integer)( numResultsBox.getSelectedItem() );
				startSearch( searchField.getText(), resultsInteger.intValue() );
				}
			}
		}
	
	/**
	 * Thread that waits for server response
	 */
	private class SearchThread extends Thread {
		
		SearchQuery sQuery;
		
		/**
		 * Constructs a <code>SearchThread</code> that will handle a search query.
		 *
		 * @param sq query to be handled by this thread
		 */
		public SearchThread( SearchQuery sq ) {
			sQuery = sq;
			setPriority( Thread.MIN_PRIORITY );
			}
		

		public void run() {
			progBar.setString( "sending query" );
			progBar.setValue( 1 );
			
			try {
				messageSock.send( sQuery );
				
				progBar.setString( "receiving response" );
				progBar.setValue( 2 );
				
				Message m = messageSock.receive();
				
				if( m instanceof SearchResults ) {
					progBar.setString( "processing results" );
					progBar.setValue( 3 );
					resultsTable.displayResults( (SearchResults) m );
					
					parent.processStats( (SearchResults) m );
					
					resetProgress();
					return;
					}
				else {
					System.out.println( "Server responded to search with an unexpected message type: " + m.getClass().toString() );
					messageSock.closeTotal();
					messageSock = null;
					resetProgress();
					connectionBroken();
					return;
					}
				}
			catch( IOException ioe ) {
				System.out.println( "Connection broken during search: " + ioe.toString() );
				messageSock = null;
				resetProgress();
				connectionBroken();
				return;
				}
			}	// end run()
		
		}	// end class SearchThread
	
	
	}