// Jason Rohrer
// UploadThread.java

/**
*
*	Thread for file upload w/ status display GUI hooks
*
*	Created 5-11-2000
*	Mods:
*		Jason Rohrer	6-2-2000	Added support for sending files from subdirectories.
*/

package konspire.client;

import konspire.common.Message;

import konspire.common.FileDescriptor;
import konspire.common.TwoWayMessageSocket;

import java.io.File;
import java.io.*;


/**
 * TransferThread for uploads.
 *
 * @author Jason Rohrer
 */
public class UploadThread extends TransferThread {
	
	int blockSize = 20000;	// 20 KB blocks
	
	File filesDir;
	
	long bytesToSkip = 0;
	
	boolean accept;
	
	TwoWayMessageSocket messageSock;
	
	Client parent;

	/**
	 * Construct an UploadThread.
	 *
	 * @param req 			file request thread will handle
	 * @param inFilesDir 	directory where files are located
	 * @param acc 			should upload request be accepted
	 * @param sock 			socket for connection
	 * @param par			parent client of upload thread
	 *
	 */	
	public UploadThread( ClientFileRequest req, File inFilesDir, boolean acc, TwoWayMessageSocket sock, Client par ) {
		super( req.getFileDescriptor() );
		filesDir = inFilesDir;
		
		bytesToSkip = req.getBytesToSkip();
		
		accept = acc;
		
		messageSock = sock;
		
		parent = par;
		}
	
	boolean aborted = false;
	
	
	// the following functions implement the TransferThread abstract functions
	
	// includes sending acceptance message.	
	protected void startTransfer() {
		
		// default status to send if download rejected
		String status = "No free download slots.";
		
		
		// first, make sure file still exists
		String filePath = f.getPath();
		
		File fileToSend = new File( filePath );
		
		if( !fileToSend.exists() || fileToSend.length() != f.getSize() ) {
			// file has changed or no longer exists
			status = "File no longer exists.";
			accept = false;
			}
		
		if( !accept ) {	
			try {
				// reject this request
				FileTransferInit init = new FileTransferInit( accept, status, blockSize );
				
				messageSock.send( init );
				messageSock.close();
				}
			catch( IOException e ) {
				System.out.println( "Connection broken while sending reject message to client: " + e.toString() );
				}
			return;
			}
			
		
		try {
			
			long fileSize = f.getSize();
			
			FileTransferInit init = new FileTransferInit( accept, "", blockSize );
				
			messageSock.send( init );
			
			
			// now start sending file
			
			byte buff[] = new byte[ blockSize ];
			
			// setup streams
			RandomAccessFile fileInput = new RandomAccessFile( filePath, "r" );
			fileInput.skipBytes( (int)bytesToSkip );
			long bytesRemain = fileSize - bytesToSkip;

			DataOutputStream output = messageSock.getOutputStream();			
			
						
			while( bytesRemain > blockSize ) { // full blocks remain
				if( aborted ) {
					messageSock.close();
					fileInput.close();
					parent.clientDisconnected();
					transferFailed();
					return;
					}
				fileInput.readFully( buff );
				output.write( buff );
				bytesRemain -= blockSize;
				setProgress( fileSize - bytesRemain );
				}
			
			// last block may be a partial block
			if( bytesRemain > 0 ) {
				if( aborted ) {
					messageSock.close();
					fileInput.close();
					parent.clientDisconnected();
					return;
					}
				fileInput.readFully( buff, 0, (int)bytesRemain );
				output.write( buff, 0, (int)bytesRemain );
				bytesRemain = 0;
				setProgress( fileSize );
				}
			
			// done uploading
			messageSock.close();
			fileInput.close();
			parent.clientDisconnected();
			return;
			
			}
		catch( IOException e ) {
			System.out.println( "Connection broken in uploading thread: " + e.toString() );
			transferFailed();
			}
			
		parent.clientDisconnected();
		return;
		}


	protected void abortTransfer() {
		aborted = true;
		}

	}