// Jason Rohrer
// TwoWayMessageSocket.java

/**
*
*	Class for sending and receiving message between hosts
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	7-3-2000	Added support for getting remote
*									host address directly from socket.
*/

package konspire.common;


import java.io.*;
import java.net.*;

/**
 * Socket for sending and receiving serialized object messages between <code>Host</code>s.
 *
 * @author Jason Rohrer
 */
public class TwoWayMessageSocket {
	
	private Socket sock;
					
	private DataOutputStream output;
	private DataInputStream input;

	private ObjectOutputStream objOutStream;
	private ObjectInputStream objInStream;

	private boolean socketCreated;	// was socket created internally (or passed in)?
	
	// host connected to other end of socket
	private Host remoteHost;
	
	
	/**
	 * Constructs a <code>TwoWayMessageSocket</code> specifying the <code>Host</code> to connect to.
	 *
	 * @param dest the <code>Host</code> to connect to
	 *
	 * @throws IOException if socket fails to initialize
	 */
	public TwoWayMessageSocket( Host dest ) 
		throws IOException {
		
		remoteHost = dest;
		
		sock = new Socket( dest.getAddress(), dest.getPort() );
		socketCreated = true;
		
		output = new DataOutputStream( sock.getOutputStream() );
		input = new DataInputStream( sock.getInputStream() );
		
		objOutStream = new ObjectOutputStream( output );
		objInStream = new ObjectInputStream( input );
		}
	
	
	/**
	 * Constructs a <code>TwoWayMessageSocket</code> from an existing
	 * established socket connection.
	 * <p>
	 * Note that socket connection WILL NOT be closed by <@link close>.<BR>
	 * Use <@link closeTotal> to close the passed in socket along with the <code>TwoWayMessageSocket</code>.
	 *
	 * @param s an already established <code>Socket</code> connection to attach
	 *  this <code>TwoWayMessageSocket</code> to.
	 *
	 * @throws IOException if attaching to socket fails
	 */
	public TwoWayMessageSocket( Socket s )
		throws IOException {
		
		sock = s;
		socketCreated = false;
		
		output = new DataOutputStream( sock.getOutputStream() );
		input = new DataInputStream( sock.getInputStream() );
		
		objOutStream = new ObjectOutputStream( output );
		objInStream = new ObjectInputStream( input );
		
		
		InetAddress remoteHostAddress = sock.getInetAddress();
		int remoteHostPort = sock.getPort();
		
		remoteHost = new Host( remoteHostAddress.getHostAddress(), remoteHostPort );
		}
	
	/**
	 * Gets the <code>Host</code> that this socket is connected to.
	 *
	 * @return the <code>Host</code> that this socket is connected to
	 */
	public Host getRemoteHost() {
		return remoteHost;
		}
	
	/**
	 * Sends a <code>Message</code> through this socket.
	 *
	 * @param m the <code>Message</code> to send
	 *
	 * @throws IOException if send fails
	 */	
	public void send( Message m ) throws IOException {
		objOutStream.writeObject( m );
		objOutStream.flush();
		} 	
	
	/**
	 * Gets the <code>DataInputStream</code> attached to this socket.
	 *
	 * @return the <code>DataInputStream</code> attached to this socket
	 */
	public DataInputStream getInputStream() {
		return input;
		}
	
	/**
	 * Gets the <code>DataOutputStream</code> attached to this socket.
	 *
	 * @return the <code>DataOutputStream</code> attached to this socket
	 */
	public DataOutputStream getOutputStream() {
		return output;
		}
	
	
	/**
	 * Receives a <code>Message</code> through this socket.
	 *
	 * @return the received <code>Message</code>
	 *
	 * @throws IOException if receive fails
	 */	
	public Message receive() throws IOException {
		try {
			Object received = objInStream.readObject();		
			if( received instanceof Message ) {
				return (Message) received;
				}
			return null;
			}
		catch( ClassNotFoundException e ) {
			System.out.println("Unknown class received on object input stream.");
			return null;
			}	
		}
	
	private boolean beenClosed = false;
	
	
	/**
	 * Closes this <code>TwoWayMessageSocket</code>.  If the corresponding
	 * <code>Socket</code> was created externally, the <code>Socket</code> is
	 * not closed, just the streams attached to it.
	 *
	 * @throws IOException if close operation fails
	 */
	public void close() throws IOException {
		if( !beenClosed ) {
			objOutStream.close();
			objInStream.close();
			output.close();
			input.close();
			
			// only close socket if created internally
			if( socketCreated ) {
				sock.close();
				}
				
			beenClosed = true;
			}
		}
	
	/**
	 * Closes this <code>TwoWayMessageSocket</code> completely, including closing 
	 * any externally created <code>Socket</code> (e.g., if <code>Socket</code>
	 * was passed in to constructor.
	 *
	 * @throws IOException if close operation fails
	 */
	public void closeTotal() throws IOException {
		socketCreated = true;
		close();
		}	
		
	}