// Jason Rohrer
// ServerPropagator.java

/**
*
*	Class for propagating messages between servers
*
*	Created 5-2-2000
*	Mods:
*/

package konspire.server;

import konspire.common.Message;
import konspire.common.Host;
import konspire.common.MessageSender;
import konspire.common.ServerListFile;

import konspire.common.ServerGoingDown;

import java.io.IOException;

import konspire.common.Vector;
import java.util.Enumeration;

/**
 * Handles propagating messages to other servers.
 *
 * @author Jason Rohrer
 */
public class ServerPropagator extends MessageSender {


	Vector hostVector = new Vector();
	
	Host self;
	
	Server parent;
	
	// for writing updated server lists out to file
	ServerListFile sListFile;
	
	/** 
	 * Constructs a propagator, specifying a list of other live hosts.
	 *
	 * @param sList list of other live <code>Host</code>s
	 * @param slf file where server list will be updated
	 * @param me the <code>Host</code> this propagator is running on (this <code>Host</code>)
	 * @param par the parent <code>Server</code> that this propagator is running on
	 */
	public ServerPropagator( Host sList[], ServerListFile slf, Host me, Server par ) {
		// deep copy
		for( int i=0; i<sList.length; i++ ) {
			// not *that* deep, just a separate list of Host pointers
			hostVector.addElement( sList[i] );
			}
		
		sListFile = slf;
		
		self = me;
		parent = par;
		hostVector.addElement( self );
		}
	
	/** 
	 * Constructs a propagator without specifying a list of known hosts.
	 *
	 * @param slf file where server list will be updated
	 * @param me the <code>Host</code> this propagator is running on (this <code>Host</code>)
	 * @param par the parent <code>Server</code> that this propagator is running on
	 */
	public ServerPropagator( ServerListFile slf, Host me, Server par ) {		
		sListFile = slf;
		
		self = me;
		parent = par;
		hostVector.addElement( self );
		}
	
	/**
	 * Adds a server to the list that this propagator is propagating message to.
	 *
	 * @param addition <code>Host</code> to add to the list
	 */
	public synchronized void addServer( Host addition ) {
		// first, make sure host vector doesn't already have this host
		
		boolean alreadyAdded = false;
		
		Enumeration e = hostVector.elements();
		while( e.hasMoreElements() && ! alreadyAdded ) {
			Object o = e.nextElement();
			if( addition.equals( o ) ) {
				alreadyAdded = true;
				}
			}
		if( ! alreadyAdded ) {	
			hostVector.addElement( addition );
			sListFile.writeHostsToFile( getHostArray() );
			System.out.println("Adding server to prop chain: " + addition.toString() );
			}
		else {
			System.out.println("Server already in prop chain: " + addition.toString() );
			}
		}
	
	
	/**
	 * Adds a vector of servers to the list that this propagator is propagating message to.
	 *
	 * @param serverVector vector of <code>Host</code>s to add to the list
	 */
	public synchronized void addServer( Vector serverVector ) {
		hostVector.addAll( serverVector );
		}
	
	
	/**
	 * Removes a server from the list that this propagator is propagating message to.
	 *
	 * @param removal <code>Host</code> to remove from the list
	 */
	public synchronized void removeServer( Host removal ) {
		hostVector.removeElement( removal );
		sListFile.writeHostsToFile( getHostArray() );
		System.out.println("Removing server from prop chain " + removal.toString() );
		}
	
	
	/**
	 * Gets the vector of <code>Host</code>s that this propagator is propagating message to.
	 *
	 * @return vector of <code>Host</code>s that this propagator is propagating message to
	 */
	public synchronized Vector getHostVector() {
		return hostVector;
		}
	

	/**
	 * Passes on a message that is already in the middle of being propagated
	 *
	 * @param p the propagation message to pass on
	 */
	public void propagateMessage( PropMessage p ) {
		
		System.out.println("Trying to propagate: " + p.getMessage().getClass().toString() );
		
		p.markServer( self );
		
		boolean sent = false;
		
		while( !sent ) {
			Host hostToSend = p.getUnmarked();
			if( hostToSend != null ) {
				
				sent = sendMessage( p, hostToSend );
				
				if( !sent  ) {	// send failed
					// server dead??
					
					System.out.println( "Message WAS NOT propagated successfully to " + hostToSend.toString() );
					
					// send message to self about dead server
					// this will cause receive thread to start propagating
					// a server dead message
					Host localAddress = parent.getLocalAddress();
					sendMessage( new ServerGoingDown( hostToSend ), localAddress );
					// this should avoid synchronization problems
					// plus, we may eventually want the server to tell clients about the
					// dead server....
					
					
					p.markServer( hostToSend );
					hostToSend = p.getUnmarked();
					}
				else {	// sent successful
					System.out.println("Propagated successfully to " + hostToSend.toString() );
					}	
				}
			else {	// no hosts left to try propagating to
				System.out.println("No hosts left to propagate message to: " + p.getMessage().getClass().toString() );
				sent = true;
				}
			}
		}
	
	// Note:  synchonized methods prevent a thread from
	// adding or removing a host from the host vector while
	// another thread is trying to start a new message
	
	
	/**
	 * Starts a propagation from a new message that hasn't been propagated yet.
	 *
	 * @param m the message to start propagating
	 */
	public void propagateMessage( Message m ) {
		// wrap message and propagate it
		PropMessage p = new PropMessage( getHostArray(), m );
		propagateMessage( p );
		}
	
	
	/**
	 * Gets array of <code>Host</code>s from host vector.
	 *
	 * @return array of <code>Host</code>s from host vector
	 */ 
	private synchronized Host [] getHostArray() {
		int size = hostVector.size();
		if( size != 0 ) {
			Host [] returnArray = new Host[ size ];
			for( int i=0; i<size; i++ ) {
				returnArray[i] = (Host)( hostVector.elementAt(i) );
				}
			return returnArray;
			}
		else {
			return null;
			}
		}
	}